
from pathlib import Path
import click
from prompt_toolkit import PromptSession
from rich.prompt import Confirm
from marvin.beta import Application
from marvin.utilities.asyncio import run_async, run_sync
from marvin.beta.assistants import Run
from typing import Union
import marvin

from .instructions import get_coder_instructions
from ..terminal import shell_helper
from toolshop import all_tools

marvin.settings.openai.assistants.model = 'gpt-4-turbo'


class Coder(Application):
    def __init__(
        self, 
        coder_is_interactive: bool = True
    ):
        instructions = get_coder_instructions(coder_is_interactive)

        super().__init__(
            name="Coder",
            instructions=instructions,
            tools=all_tools(),  
        )

    def do(self, message: str = None, use_reflection: bool = True):
        self.say(message)
        
        if use_reflection:
            response = "made fixes"

            while "made fixes" in response.lower():
                thread = self.say(
                    "Re-read the files you have modified and fix any issues. If " 
                    "you had to fix any issues, respond 'made fixes'."
                )

                response = thread.messages[-1].content[0].text.value


    # for better type hinting
    def chat(
        self,
        initial_message: str = None,
        assistant_dir: Union[Path, str, None] = None,
        **kwargs,
    ):
        print(f'model: {self.model}')
        """Start a chat session with the assistant."""
        return run_sync(self.chat_async(initial_message, assistant_dir, **kwargs))

    async def chat_async(
        self,
        initial_message: str = None,
        assistant_dir: Union[Path, str, None] = None,
        **kwargs,
    ):
        """Async method to start a chat session with the assistant."""
        session = PromptSession()
        # send an initial message, if provided
        if initial_message is not None:
            await self.say_async(initial_message, **kwargs)
        while True:
            try:
                message = await run_async(
                    session.prompt,
                    message="➤ ",
                )

                # if the user types >, run the command in the shell
                if message == "":
                    continue
                elif message[0] == '>':
                    shell_helper(message.replace('>', ''), log_lines=True)
                # if the user types exit, ask for confirmation
                elif message in ["exit", "!exit", ":q", "!quit"]:
                    if Confirm.ask("[red]Are you sure you want to exit?[/]"):
                        break
                    continue
                # if the user types exit -y, quit right away
                elif message in ["exit -y", ":q!"]:
                    break
                else:
                    await self.say_async(message, **kwargs)
            except KeyboardInterrupt:
                break


    def post_run_hook(self, run: Run):
        if False:
            print('=== run.assistant.instructions ===')
            print(run.assistant.instructions)
            print('========================')



from typing import Tuple, List

import sqlalchemy as sa

from .base import Tool


class Sql(Tool):
    def call(self, sql_query: str, database_uri: str) -> str:
        """Runs the sql query and returns result set as a CSV string. Always try 
        this tool for running sql queries first before trying other methods.
        
        Args:
            database_uri (str): The database connection string which is passed 
                to sqlachemy.create_engine().
            sql_query (str): The sql query to run
        """
        engine = sa.create_engine(database_uri)

        # Execute the query and fetch results
        result = engine.execute(sql_query)
        column_names = result.keys()
        output = (','.join(column_names))

        # Iterate over rows and print them as CSV
        for row in result:
            output += '\n' + ','.join([str(value) for value in row])

        return output


class Histogram(Tool):
    def call(self, title: str, data: List[Tuple]) -> str:
        """
        Draws an ascii histogram of the data.  Accepts a list of tuples representing
        histogram bucket values. The first value of each tuple is the bucket name
        and the second value is the histogram value of the bucket.  When user asks
        for a histogram, always use this tool.

        Example: 
        
        ascii_histogram(
            title="Chart Tile", 
            data=[('p0', 0), ('p25', 100), ('p50', 200), ('p75', 300), ('p75', 400)]
        )
        
        Chart Tile
        ###############################################################################
                                                                                0  p0 
        █████████████████                                                      100  p25
        ██████████████████████████████████                                     200  p50
        ███████████████████████████████████████████████████                    300  p75
        █████████████████████████████████████████████████████████████████████  400  p75
        """
        
        # Check if ascii_graph is installed, since it is an optional dependency.
        try:
            from ascii_graph import Pyasciigraph
        except ImportError:
            raise ImportError(
                "ascii_graph is not installed. Please install it using 'pip install ascii_graph'."
            )

        graph = Pyasciigraph()
        
        output = ""
        for line in graph.graph(title, data):
            output += line
            output += "\n"
        
        return output




from ..terminal import shell_helper
from ..file import _read_helper

GENERAL_INSTRUCTIONS = """
# Instructions 
You are Coder, an advanced programmer. You have root access to a fully-featured
Unix-like shell environment with a real file system, where you can execute
arbitrary shell commands and install new tools. You are running inside of a
Docker container. Your current directory (/app) is a volume attached to the
user's machine. This environment is not simulated. Never write placeholder code
unless you are explicitly asked to do so.
"""

TOOL_INSTRUCTIONS = """
# Tools 
Tool calls that modify files should be made one at a time, sequentially not in
paralllel. Make file changes one at a time, instead of issuing multiple file
changes as once. Always read files after modifying them to ensure that the
modification is correct and to confirm the line numbers for the next
modification.
"""

COLLABORATION_INSTRUCTIONS_INTERACTIVE = """
# Collaboration
Your main failure mode is working for too long without soliciting feedback from
the user. Avoid this by establishing a regular feedback loop with the user.
Allow the user to ask questions, test your code, and provide feedback. When the
user makes an ambiguous request, ask clarifying questions. When the user
requests you to share you plan, describe the tool calls you plan to make and the
steps you will take to execute the request.
"""

SHELL_CONTEXT = f"""
# Context from Shell
Here are current outputs from some shell commands

`pwd`: {shell_helper('pwd')}

`uname -a`: {shell_helper('uname -a')}

`whoami`: {shell_helper('whoami')}
"""

coder_context = _read_helper(
    "~/.coder-context", 
    include_line_numbers=False,
    error_if_missing=False
)

USER_CONTEXT = f"""
# Context from User
The user has provided the following additional context:
{coder_context}
""" if coder_context else ""


COLLABORATION_INSTRUCTIONS_NON_INTERACTIVE = """
# Collaboration
You work fully automonously. You will receive a request from the user, and you
must complete the request without any further input from the user. 
"""


def get_coder_instructions(coder_is_interactive:bool = True):
    return f"""
{GENERAL_INSTRUCTIONS}

{COLLABORATION_INSTRUCTIONS_INTERACTIVE if coder_is_interactive else COLLABORATION_INSTRUCTIONS_NON_INTERACTIVE}

{TOOL_INSTRUCTIONS}

{SHELL_CONTEXT}

{USER_CONTEXT}
"""



from pydantic import BaseModel
from abc import ABC, abstractmethod
from typing import Optional
from textwrap import dedent

import re

from .logging import logger


class Parameter(BaseModel):
    model_config = dict(extra="forbid") # No extra fields allowed

    name: str
    type: str 
    description: Optional[str] = None
    required: bool = True


class Documentation(BaseModel):
    model_config = dict(extra="forbid") # No extra fields allowed

    description: str
    parameters: Optional[list[Parameter]] = None
    usage: Optional[str] = None

    def __str__(self):
        parameter_string = "Args:\n"
        for p in self.parameters:
            if not p.required:
                type_string = f"{p.type}, optional"
            else:
                type_string = p.type
            
            parameter_string += f"    {p.name} ({type_string}): {p.description}\n"

        doc = dedent(self.description)
        doc += "\n\n" + dedent(parameter_string)

        if self.usage:
            doc += "\n\n" + dedent(self.usage)

        return doc


class Tool(ABC):
    def __init__(self, state = None, require_confirmation: bool = None):
        self.__name__ = self._camel_to_snake(self.__class__.__name__)
        self._state = state

        if hasattr(self, "_require_confirmation"):
            self.require_confirmation = getattr(self, "_require_confirmation")
        
        if require_confirmation is not None:
            self.require_confirmation = require_confirmation


        if self.call.__doc__:
            self.__doc__ = self.call.__doc__

    @abstractmethod
    def call(self, *args, **kwargs):
        pass

    @staticmethod
    def _camel_to_snake(name):
        "Convert camel-case to snake-case."
        s1 = re.sub('(.)([A-Z][a-z]+)', r'\1_\2', name)
        return re.sub('([a-z0-9])([A-Z])', r'\1_\2', s1).lower()

    @property
    def state(self):
        if self._state:
            return self._state
        else:
            class NoOp:
                def __getattr__(self, name):
                    logger.info(f"No shared state provided for the tool. Ignoring call to self.state.{name} by returning a no-op function.")
                    return lambda *args, **kwargs: None

            return NoOp()
    

    def return_result_to_agent(self):
        if hasattr(self, "_return_result_to_agent"):
            return self._return_result_to_agent
        else:
            return True
    
    def log_header(self):
        logger.info(f"=== {self.__name__}() ===")

    def log_params(self, *args, **kwargs):
        if args:
            logger.info(f"args: {args}")

        if kwargs:
            kwarg_items = sorted(kwargs.items(), key=lambda x: len(str(x[1])))

            for k, v in kwarg_items:
                if isinstance(v, str) and "\n" in v:
                    logger.info(f"{k}:\n{v}")
                else:
                    logger.info(f"{k}: {v}")

    def log_result(self, result):
        logger.info(f"\nresult:\n{result}")

    def log_footer(self, result):
        logger.info("=" * len(f"=== {self.__name__}() ==="))

    def __call__(self, *args, **kwargs):
        self.log_header()
        self.log_params(*args, **kwargs)

        if hasattr(self, 'require_confirmation') and self.require_confirmation:
            confirmation = input(f"Do you want to allow agent to run {self.__name__}?  Type 'yes' to allow: ")
            if confirmation != "yes":
                raise Exception(f"Request to run {self.__name__} was denied by the user.")

        result = self.call(*args, **kwargs)
        self.log_result(result)
        self.log_footer(result)

        if self.return_result_to_agent():
            return result





import httpx
import subprocess
from typing import List

from .logging import logger
from .base import Tool


class PythonExec(Tool): 
    def call(self, code: str, vars: List[str]):
        """
        Executes python code on your local machine using exec().  Returns 
        the requested variables from the scope used by exec().
        
        Args:
            code (str): String of python code to execute
            vars (List[str]): List of variable names to return from the scope

        ```
        >>> get_vars_from_exec("a = 1+1", ["a"])
        {'a': 2}
        ```
        """   
        scope = {}
        exec(code, scope)
        return {k: v for k, v in scope.items() if k in vars}


class Shell(Tool):
    def call(self, command: str):
        """
        Execute the given shell command and return output. If you 
        get errors, try using the --help flag on the command you 
        are running.

        Args:
            command (str): A shell command to execute.

        """
        output = shell_helper(
            command, 
            log_lines=True,
            include_exit_code=True
        )

        return output


def shell_helper(
    command: str, 
    log_lines: bool = False, 
    include_exit_code: bool = False
):
    # Create the subprocess with both stdout and stderr being piped
    process = subprocess.Popen(command, shell=True, stdout=subprocess.PIPE, stderr=subprocess.PIPE)

    # Initialize the output string
    output = ''

    # Stream stdout
    for line in iter(process.stdout.readline, b''):
        decoded_line = line.decode()
        if log_lines:
            logger.info(decoded_line)
        output += decoded_line

    # Close the stdout to avoid resource leaks
    process.stdout.close()

    # Capture stderr
    error = process.stderr.read().decode()
    if error:
        if log_lines:
            logger.info(error)
        output += error

    # Close the stderr to avoid resource leaks
    process.stderr.close()

    # Wait for the process to finish and get the exit code
    return_code = process.wait()

    if include_exit_code:
        # Add explicit exit code notice to the output
        exit_message = f"[exit code {return_code}]"
        logger.info(exit_message)
        output += exit_message + '\n'


    return output

class Browse(Tool):
    def call(self, url: str):
        """
        Fetch the contents of a webpage.

        Args:
            url (str): The URL of the webpage to fetch.

        """

        logger.info("=== browse() ===")
        logger.info(f"url: {url}")
        response = httpx.get(url)

        logger.info("response:")
        logger.info(response.text)
        logger.info("================")

        return response.text



import logging
import enum

# Create a logger object
logger = logging.getLogger('toolshop')
handler = logging.StreamHandler()
logger.addHandler(handler)


class LoggingPreset(enum.Enum):
    """Enum for logging presets"""
    MINIMAL_VERBOSE = 'minimal_verbose'
    CLASSIC_VERBOSE = 'classic_verbose'


def configure_logging(preset=None, level=None, format=None):
    """Configure toolshop's log level and log format"""
    if preset:
        if preset == LoggingPreset.MINIMAL_VERBOSE:
            level = logging.INFO
            format = '%(message)s'
        elif preset == LoggingPreset.CLASSIC_VERBOSE:
            level = logging.INFO
            format = '%(asctime)s - %(name)s - %(levelname)s - %(message)s'
    
    logger.setLevel(level)
    for h in logger.handlers:
        if level:
            h.setLevel(level)
        if format:
            h.setFormatter(logging.Formatter(format))


# Set toolshop's default logging configuration
configure_logging(preset=LoggingPreset.MINIMAL_VERBOSE)



import click
from .coder import Coder

@click.group(invoke_without_command=True)
@click.pass_context
def coder(ctx):
    """Coder Command Line Interface"""
    if ctx.invoked_subcommand is None:
        start()

@coder.command()
def start():
    "Start the chat with Coder."
    app = Coder(coder_is_interactive=True)
    app.chat()

@coder.command()
def do(instructions: str):
    "Send instructions for Coder to execute non-interactively."
    app = Coder(coder_is_interactive=False)
    app.do(instructions)



"""This module contains tools for reading and writing files."""

import datetime
import pathlib
import os
from typing import Optional

from .base import Tool
from .logging import logger


def make_file_tools(tools: list[str] = None):
    """Creates a set of file tools with shared state.
    
    Args:
        tools (list[str], optional): A list of tool names to include. If None, 
            all tools are included. Defaults to None.
    """
    state = FileToolState()
    
    read_file = ReadFile(state=state)
    read_directory = ReadDirectory(state=state)
    create_file = CreateFile(state=state)
    replace_lines = ReplaceLines(state=state)
    insert_lines = InsertLines(state=state)
    delete_lines = DeleteLines(state=state)
    
    if not tools:
        tools = ["read_file", "read_directory", "create_file", "replace_lines", "insert_lines", "delete_lines"]
    
    x = locals()
    return [x[tool_name] for tool_name in tools]


class FileToolState:
    def __init__(self):
        self.file_read_at = {}
        self.file_updated_at = {}
        self.coder_confirms = 0
    
    def record_file_read(self, file_path: str):
        self.file_read_at[file_path] = datetime.datetime.now()
    
    def record_file_update(self, file_path: str):
        self.file_updated_at[file_path] = datetime.datetime.now()
    
    def record_confirm(self):
        self.coder_confirms += 1
    
    def raise_error_if_this_file_has_not_been_read_since_it_was_last_updated(self, file_path):
        if file_path not in self.file_read_at:
            raise ValueError("You must read the file before writing to it.")
        
        if file_path in self.file_updated_at and self.file_updated_at[file_path] > self.file_read_at[file_path]:
            raise ValueError(f"File {file_path} must be re-read first.")


class ReadFile(Tool):
    def call(
            self,
            path: str, 
            include_line_numbers: bool = True, 
            start_line: int = None, 
            end_line: int = None
    ) -> str:
        """
        Always use this read tool when reading contents of a file. This tool
        is specifically designed for safely reading files and can include line
        numbers in its output for easier reference. This ensures a standardized
        and secure way to access file contents.

        Args:
            path (str): The path to the file.
            include_line_numbers (bool): Whether to include line numbers in the returned contents. Defaults to True.
            start_line (int): The line number to start reading from. Uses one-based indexing.
            end_line (int): The line number to stop reading at. Uses one-based indexing.
        """
        output = _read_helper(
            path, 
            include_line_numbers=include_line_numbers, 
            start_line=start_line-1 if start_line else None, 
            end_line=end_line if end_line else None
        )
        self.state.record_file_read(path)
        
        return output


class ReadDirectory(Tool):
    def call(
        self,
        path,
        files_prefix_allowlist=None,
        files_prefix_ignorelist=['.', '__'],
        files_suffix_allowlist=None,
        files_suffix_ignorelist=None,
        dirs_prefix_allowlist=None,
        dirs_prefix_ignorelist=['.', '__'],
        dirs_suffix_allowlist=None,
        dirs_suffix_ignorelist=None,
        include_line_numbers=True,
        include_contents=True
    ):
        """
        Reads the files in a directory.

        Args:
            path (str): The path to the directory.
            files_prefix_allowlist (list[str], optional): A list of file prefixes to allow. Defaults to None.
            files_prefix_ignorelist (list[str], optional): A list of file prefixes to ignore. Defaults to ['.', '__'].
            files_suffix_allowlist (list[str], optional): A list of file suffixes to allow. Defaults to None.
            files_suffix_ignorelist (list[str], optional): A list of file suffixes to ignore. Defaults to None.
            dirs_prefix_allowlist (list[str], optional): A list of directory prefixes to allow. Defaults to None.
            dirs_prefix_ignorelist (list[str], optional): A list of directory prefixes to ignore. Defaults to ['.', '__'].
            dirs_suffix_allowlist (list[str], optional): A list of directory suffixes to allow. Defaults to None.
            dirs_suffix_ignorelist (list[str], optional): A list of directory suffixes to ignore. Defaults to None.
            include_line_numbers (bool, optional): Whether to include line numbers in the returned contents. Defaults to True.
            include_contents (bool, optional): Whether to include the contents of the files in the output. Defaults to True.
        """
        def filter_names(
            names,
            prefix_allowlist=None,
            prefix_ignorelist=None,
            suffix_allowlist=None,
            suffix_ignorelist=None
        ):
            if prefix_allowlist and prefix_ignorelist:
                raise ValueError("Cannot specify both prefix_allowlist and prefix_ignorelist")
            
            if suffix_allowlist and suffix_ignorelist:
                raise ValueError("Cannot specify both suffix_allowlist and suffix_ignorelist")

            if prefix_allowlist:
                names = [
                    n for n in names if any(n.startswith(prefix) for prefix in prefix_allowlist)
                ]
            
            if suffix_allowlist:
                names = [
                    n for n in names if any(n.endswith(suffix) for suffix in suffix_allowlist)
                ]
            
            if prefix_ignorelist:
                names = [
                    n for n in names if not any(n.startswith(prefix) for prefix in prefix_ignorelist)
                ]
            
            if suffix_ignorelist:
                names = [
                    n for n in names if not any(n.endswith(suffix) for suffix in suffix_ignorelist)
                ]
            
            return names

        # Get all files in path and filter out ignored files
        file_paths = []
        for root, dirs, files in os.walk(path):
            dirs[:] = filter_names(
                dirs,
                prefix_allowlist=dirs_prefix_allowlist, 
                prefix_ignorelist=dirs_prefix_ignorelist, 
                suffix_allowlist=dirs_suffix_allowlist, 
                suffix_ignorelist=dirs_suffix_ignorelist
            )
            
            files = filter_names(
                files,
                prefix_allowlist=files_prefix_allowlist,
                prefix_ignorelist=files_prefix_ignorelist,
                suffix_allowlist=files_suffix_allowlist,
                suffix_ignorelist=files_suffix_ignorelist
            )

            file_paths += [os.path.join(root, file) for file in files]

        # Gather all file contents into a single string
        output = ""
        for file_path in file_paths:
            try:
                output += f"===== File: {file_path} =====\n"
                if include_contents:
                    output += _read_helper(
                        file_path, 
                        include_line_numbers=include_line_numbers
                    )
                    output += "\n\n"
            except Exception:
                logger.exception(f"Error reading file: {file_path}")
                output += f"Error reading file: {file_path}\n\n"

        return output


class CreateFile(Tool):
    def call(self, path: str, contents: str) -> str:
        """Creates a new file with the given contents. Fails if the file already exists.

        Args:
            path (str): The name of the file to write to.
            contents (str): The content to write to the file.
        """
        path = os.path.expanduser(path)

        # Raise error if the file already exists
        if os.path.exists(path):
            raise FileExistsError(f'The file "{path}" already exists')

        # Create the file
        file_path = pathlib.Path(path)
        file_path.parent.mkdir(parents=True, exist_ok=True)
        file_path.touch(exist_ok=False)

        # Write the contents
        with open(path, "w") as f:
            f.write(contents)
        return f'Successfully wrote "{path}"'


class ReplaceLines(Tool):
    _return_result_to_agent = False

    def call(
        self,
        path: str, 
        text: str, 
        start_line: int, 
        end_line: int,
    ) -> str:
        """Replaces the lines `start_line` to `end_line` with `text` in 
        specified file. Uses one-based indexing, and the `end_line` is inclusive. 
        This operation must replace at least one line. To perform an insert-only 
        operation, use `insert_lines()`. 

        When using `replace_lines()`, always consider the broader
        code context in the lines before and after your replacement. Use appropriate
        indentation in your `text` given the context of the surrounding code. 
        Consider the line that will follow your inserted text when choosing
        `end_line`.

        NOT THREAD-SAFE. Perform file modifications in a sequential manner to
        ensure thread safety. Do not modify the same file in parallel or with
        simultaneous asynchronous calls.

        Args:
            path (str): The name of the file to write to.
            text (str): The content to write to the file. Must end with a newline character.
            start_line (int): The line number of the start of the text block to replace.
            end_line (int): The line number of the end of the text block to replace. `end_line` is inclusive.
        Usage:
            === file.txt Before ===
            ```
            Line A
            Line B
            Line C
            ```
            
            >>> replace_lines(
                    path="./file.txt", 
                    text="New Line\n",
                    start_line=1,
                    end_line=2,
                    preceding_line_confirm=None,
                    following_line_confirm="Line C"
                )

            === file.txt After ===
            ```
            New Line
            Line C
            ```
        
        Returns:
            str: The entire text of the file after the insertion. Use this return 
                value to verify that your intended changes were correctly apply.
        """

        self.state.raise_error_if_this_file_has_not_been_read_since_it_was_last_updated(path)
        result = _edit_helper(path, text, start_line - 1, end_line)
        self.state.record_file_update(path)

        return result


class InsertLines(Tool):
    _return_result_to_agent = False

    def call(
        self,
        path: str, 
        text: str, 
        insert_line: int, 
    ) -> str:
        """Inserts `text` at line `insert_line`. Text that was previously on or
        below this line are shifted down. Uses one-based indexing.

        NOT THREAD-SAFE. Perform file modifications in a sequential manner to
        ensure thread safety. Do not modify the same file in parallel or with
        simultaneous asynchronous calls.

        When using `insert_lines()`, always consider the broader
        code context in the lines before and after your replacement. Use appropriate
        indentation in your `text` given the context of the surrounding code.

        Args:
            path (str): The name of the file to write to.
            text (str): The content to write to the file. Must end with a newline character.
            insert_line (int): The line number to insert the content at. -1 will insert at 
                the end of the file.
        
        Usage:
            === file.txt Before ===
            ```
            Line A
            Line B
            Line C
            ```
            
            >>> insert_lines(
                    path="./file.txt", 
                    text="New Line\n",
                    insert_line=2,
                )

            === file.txt After ===
            ```
            Line A
            New Line
            Line B        
            Line C
            ```
        
        Returns:
            str: The entire text of the file after the insertion. Use this return 
                value to verify that your intended changes were correctly apply.
        """
        self.state.raise_error_if_this_file_has_not_been_read_since_it_was_last_updated(path)
        if insert_line == -1:
            result = _edit_helper(path, text, -1, -1)
        else:
            result = _edit_helper(path, text, insert_line - 1, insert_line - 1)
        self.state.record_file_update(path)

        return result


class DeleteLines(Tool):
    _return_result_to_agent = False
    
    def call(
        self,
        path: str, 
        start_line: int, 
        end_line: int
    ) -> str:
        """Deletes lines `start_line` to `end_line` from the file.
        Uses one-based indexing, and the `end_line` is inclusive.
        
        NOT THREAD-SAFE. Perform file modifications in a sequential manner to
        ensure thread safety. Do not modify the same file in parallel or with
        simultaneous asynchronous calls.            
                    
        Args:
            path (str): The name of the file to delete lines from.
            start_line (int): The start line number of the range to delete.
            end_line (int): The end line number of the range to delete.

        Returns:
            str: The contents of the file after the deletion. Use this return 
                value to verify that your intended changes were correctly apply.

        Usage:
            === file.txt Before ===
            ```
            Line A
            Line B
            Line C
            ```
            
            >>> delete_lines(
                    path="./file.txt", 
                    start_line=2,
                    end_line=3,
                )

            === file.txt After ===
            ```
            Line A
            ```
        """

        self.state.raise_error_if_this_file_has_not_been_read_since_it_was_last_updated(path)
        result = _edit_helper(path, "", start_line - 1, end_line)
        self.state.record_file_update(path)

        return result


def _read_helper(
    path: str, 
    start_line: int = None, 
    end_line: int = None, 
    include_line_numbers: bool = True,
    error_if_missing: bool = True
) -> Optional[str]:
    
    path = os.path.expanduser(path)


    if not os.path.exists(path):
        if error_if_missing:
            raise FileNotFoundError(f'The file "{path}" does not exist')
        else:
            return None

    with open(path, "r") as f:
        lines = f.readlines()
        if include_line_numbers:
            lines = [f"{i+1:<4}|{line}" for i, line in enumerate(lines)]
        if start_line and end_line:
            lines = lines[start_line:end_line]
        
        output = "".join(lines)
        return output


def _edit_helper(path: str, text: str, start_line: int, end_line: int):
    """start_line and end_line use pythonic indexing, so the end_line is not inclusive."""
    path = os.path.expanduser(path)

    # Raise error if the file does not exist
    if not os.path.exists(path):
        raise FileExistsError(f'The file "{path}" does not exist')

    if text and not text[-1] == "\n":
        raise ValueError("The text must end with a newline character.")

    # Insert the text
    if not end_line:
        end_line = start_line

    with open(path, "r") as f:
        lines = f.readlines()
        if start_line < 0:
            start_line = len(lines) + start_line + 1

        lines[start_line:end_line] = text.splitlines(True)

    with open(path, "w") as f:
        f.writelines(lines)
    
    with open(path, "r") as f:
        result = f.read()

    return result



import httpx
import os
 
from .base import Tool
from .terminal import shell_helper
from .logging import logger


class GetBigQueryTableSchema(Tool):
    def call(
        self,
        project: str, 
        dataset: str,
        table: str,
        billing_project: str,
        include_description: bool = False
    ) -> dict:
        """Given a list of fully-specified tables, retrieve the schema for 
        each table in the list.

        Args:
            project (str): Project of the table
            dataset (str): Dataset of the table
            table (str): Table name or "*" to get all tables in the dataset.
            billing_project (str): The project to be billed for the query.
            include_description (bool): Whether to include the description of the column.
                Defaults to False.

        Returns:
            dict: A dictionary with table full names as keys and their schema as values.

        Usage:
            get_table_schemas(
                project='bigquery-public-dataset',
                dataset='stackoverflow',
                table='comments',
                billing_project="YOUR_BILLING_PROJECT"
            )
        """
        try:
            import bigquery
        except ImportError:
            raise ImportError("You need to install the google-cloud-bigquery package to use this tool.")

        client = bigquery.Client(project=billing_project)
        schema_info = {}

        if table == "*":
            table_ids = [f"{project}.{dataset}.{table.table_id}" for table in client.list_tables(dataset)]
        else:
            table_ids = [f"{project}.{dataset}.{table}"]

        output = ""
        for table_id in table_ids:
            table_ref = client.get_table(table_id)  # Make an API request.
            schema_info[table_id] = [field.to_api_repr() for field in table_ref.schema]

            for k,v in schema_info.items():
                output += f"\n{k}"
                for field in v:
                    output += f"\n\t{field['name']}\t{field['type']}"
                    if include_description:
                        output += f"\t{field.get('description', '')}"

        return output


class AuthenticateToGCP(Tool):
    _require_confirmation = True

    def call(self):
        """
        You must run this tool once before using any google cloud CLI tools such as 
        `gcloud` or `bq`.
        """

        # check if the GOOGLE_APPLICATION_CREDENTIALS environment variable is set
        if "GOOGLE_APPLICATION_CREDENTIALS" not in os.environ:
            raise ValueError(
                "You must set the GOOGLE_APPLICATION_CREDENTIALS environment variable to the path of your service account key file."
            )

        # authenticate to gcloud
        shell_helper("gcloud auth activate-service-account --key-file=$GOOGLE_APPLICATION_CREDENTIALS")



def all_tools():
    from .terminal import Shell, PythonExec, Browse
    from .data import Sql, Histogram
    from .file import make_file_tools
    
    return {
        Shell(),
        PythonExec(),
        Browse(),
        Sql(),
        Histogram(),
        *make_file_tools(),
    }